from .config import *
from .chat import *
from .conversations import *
from .encryption import *
from .printer import *
from .main import *

def one_shot_message(ctx, question):
    chat_manager: ChatManager = ctx.obj["CHAT"]
    enc_manager: EncryptionManager = ctx.obj["ENC_MNGR"]
    printer: Printer = ctx.obj["PRINTER"]

    enc_manager.set_api_key()

    messages = [config.INIT_SYSTEM_MESSAGE]

    messages.append({"role": "user", "content": question})

    printer.printt("")
    answer = chat_manager.get_user_answer(messages=messages)
    message = answer["choices"][0]["message"]["content"]

    printer.print_assistant_message(message)



"""TerminalGPT configuration file."""

import json
import platform
from os import path

APP_NAME = "terminalgpt-pro"
DEFAULT_API_TOKEN_LIMIT = 4096

BASE_PATH = f"~/.{APP_NAME}".replace("~", path.expanduser("~"))
DEFAULTS_PATH = f"{BASE_PATH}/defaults.json"
CONVERSATIONS_PATH = f"{BASE_PATH}/conversations"
SECRET_PATH = f"{BASE_PATH}/{APP_NAME}.encrypted"
KEY_PATH = f"{BASE_PATH}/{APP_NAME}.key"

ENCODING_MODEL = "cl100k_base"

MODELS = {
    "gpt-3.5-turbo": 4097,
    "gpt-3.5-turbo-16k": 16385,
    "gpt-4": 8192,
    "gpt-4-32k": 32768,
}


def machine_info():
    """Get the current machine info."""

    return platform.platform()


def get_default_config() -> dict:
    """Get the default configuration from the config file."""
    try:
        with open(DEFAULTS_PATH, "r", encoding="utf-8") as file:
            return json.load(file)
    except FileNotFoundError:
        return {"model": "gpt-3.5-turbo", "style": "markdown"}
    except ValueError:
        return {"model": "gpt-3.5-turbo", "style": "markdown"}

INIT_SYSTEM_MESSAGE = {
    "role": "system",
    "content": f"""
- Your name is "TerminalGPT-Pro".
- You are a helpful personal assistant for programers.
- You are running on {machine_info()} machine.
- Please note that your answers will be displayed on the terminal.
- So keep answers short as possible and use a suitable format for printing on a terminal.
""",
}


INIT_WELCOME_MESSAGE = {
    "role": "system",
    "content": """
- Please start the conversation with a random and short greeting message starts with 'Welcome to TerminalGPT-Pro'.
- Add a ton of self humor.
- Keep it short as possible, one line.
""",
}

INIT_WELCOME_BACK_MESSAGE = {
    "role": "system",
    "content": """
The conversation you remember was a while ago, now we are continuing it.
Please start the conversation with a random and short welcome back message.
- Start with 'Welcome back to TerminalGPT-Pro'.
- Add a ton of self humor.
- Keep it short as possible, one line.

After the welcome back message, please summarize the last conversation. (e.g. "Last time we talked about ...")
- End with a something that invites the user to continue the conversation.
""",
}

TITLE_MESSAGE = """
Please give this conversation a short title.
I'm going to use this title as a file name for the conversation.
There are going to a lot of files like that under a folder "~/.terminalgptpro/conversations"
- Hard limit of 5 words.
- Use underscores instead of spaces.
- Don't mention yourself in it. (e.g. "TerminalGPT-Pro conversation")
- Don't use any special characters.
- Don't use any numbers.
- Don't use any capital letters.
- Don't use any spaces.
- Don't use any punctuation.
- Don't use any symbols.
- Don't use any emojis.
- Don't use any accents.
- Don't use quotes.
- Don't use word like: "macos", "programmer_assistant", "conversation".
- Don't use any file extensions. (e.g. ".txt" or ".json")
"""



import os
import time
import getpass
import json
from .config import *
from .chat import *
from .conversations import *
from .encryption import *
from .printer import *
from .main import *


def install_api_key():
    """Install the terminalgpt openai api key and create app directories."""

    printer: Printer = PrinterFactory.get_printer(style="plain")
    enc_manager: EncryptionManager = EncryptionManager()

    printer.printt(PrintUtils.INSTALL_WELCOME_MESSAGE)

    printer: Printer = PrinterFactory.get_printer(style="plain")
    enc_manager: EncryptionManager = EncryptionManager()

    printer.printt(PrintUtils.INSTALL_WELCOME_MESSAGE)

    while True:
        api_key = getpass.getpass(
            prompt=f"{Style.RESET_ALL}{Style.BRIGHT}Please enter your OpenAI API key:\n{Style.RESET_ALL}"
        )

        printer.printt(f"{Style.BRIGHT}{Fore.GREEN}Great!{Style.RESET_ALL}\n")
        time.sleep(0.5)

        models = list(config.MODELS.keys())
        printer.printt(
            f"{Style.BRIGHT}Please choose one of the models below to be your default model:"
        )

        for model, token_limit in config.MODELS.items():
            printer.printt(
                f"{Style.BRIGHT} - Model: {Style.RESET_ALL}{model}"
                f"{Style.BRIGHT}    tokens-limit: {Style.RESET_ALL}{int(int(token_limit)/1000)}k"
            )

        model = prompt(
            "\nType the desired model:\n",
            completer=WordCompleter(models, ignore_case=True),
            style=PromptStyle.from_dict({"prompt": "bold lightblue"}),
            default=models[0],
        )

        printing_styles = ["markdown", "plain"]
        printer.printt(
            f"{Style.BRIGHT}Please choose one of the printing styles below to be your default printing style:"
        )
        for style in printing_styles:
            printer.printt(f"{Style.BRIGHT} - Style: {Style.RESET_ALL}{style}")

        printing_style = prompt(
            "\n Choose a printing style ('markdown' is recommended):\n",
            completer=WordCompleter(printing_styles, ignore_case=True),
            style=PromptStyle.from_dict({"prompt": "bold lightblue"}),
            default=printing_styles[0],
        )

        if model in models and printing_style in printing_styles:
            printer.printt(f"\n{Style.BRIGHT}{Fore.GREEN}Great!{Style.RESET_ALL}\n")
            break
        else:
            invalid_input = ""
            if model not in models:
                invalid_input += "Invalid model! Please choose from the provided list.\n"
            if printing_style not in printing_styles:
                invalid_input += "Invalid printing style! Please choose from the provided list.\n"

            printer.printt(f"\n{Style.BRIGHT}{Fore.RED}{invalid_input}{Style.RESET_ALL}\n")
            continue

    encryption_key = enc_manager.set_encryption_key()
    encrypted_secret = enc_manager.encrypt(api_key.encode(), encryption_key)

    if not os.path.exists(os.path.dirname(config.SECRET_PATH)):
        os.makedirs(os.path.dirname(config.SECRET_PATH))

    if not os.path.exists(config.CONVERSATIONS_PATH):
        os.mkdir(config.CONVERSATIONS_PATH)

    # Save the encrypted secret to a file
    with open(config.SECRET_PATH, "wb") as file:
        file.write(encrypted_secret)

    # Save the default config to a file
    with open(config.DEFAULTS_PATH, "w", encoding="utf-8") as file:
        json.dump({"model": model, "style": printing_style}, file)

    printer.printt(PrintUtils.INSTALL_SUCCESS_MESSAGE)
    printer.printt(PrintUtils.INSTALL_ART)
    printer.printt(PrintUtils.INSTALL_SMALL_PRINTS)



"""This module contains the Printer class, which is responsible for printing messages to the user."""

import random
import time
from abc import ABC, abstractmethod
from typing import List

from colorama import Fore, Style
from rich.console import Console
from rich.markdown import Markdown

from terminalgptpro import config


class Printer(ABC):
    PRINT_DELAY = 0.006

    @abstractmethod
    def printt(self, text: str = ""):
        pass

    def print_assistant_message(self, message, color=Fore.YELLOW):
        pass


class PlainPrinter(Printer):
    def printt(self, text: str = ""):
        try:
            for char in text:
                print(char, end="", flush=True)
                time.sleep(self.PRINT_DELAY)
        except KeyboardInterrupt:
            print()
        finally:
            print()

    def print_assistant_message(self, message, color=Fore.YELLOW):
        """Prints the assistant message."""
        print(Style.BRIGHT + "TerminalGPT-Pro:" + Style.RESET_ALL)
        self.printt(color + message + Style.RESET_ALL)


class MarkdownPrinter(Printer):
    def __init__(self):
        self.__console = Console()

    def printt(self, text: str = "", style="yellow"):
        text_markdown = ""
        txt_arr = self._split_highlighted_string(text)
        for txt in txt_arr:
            with self.__console.capture() as capture:
                text_markdown = Markdown(txt)
                self.__console.print(text_markdown, style=style)
            text_markdown = capture.get()
            if txt.startswith("```"):
                delay = self.PRINT_DELAY / 10
            else:
                delay = self.PRINT_DELAY
            self._print_with_delay(text_markdown, delay)

    def print_assistant_message(self, message, color=Fore.YELLOW):
        """Prints the assistant message."""
        print(Style.BRIGHT + "TerminalGPT-Pro:" + Style.RESET_ALL)
        self.printt(color + message + Style.RESET_ALL)

    def _split_highlighted_string(self, string: str) -> List[str]:
        result = []
        start = 0
        while True:
            start_block = string.find("```", start)
            if start_block == -1:
                result.append(string[start:])
                break
            end_block = string.find("```", start_block + 3)
            if end_block == -1:
                result.append(string[start:])
                break
            result.append(string[start:start_block])
            result.append(string[start_block : end_block + 3])
            start = end_block + 3
        return result

    def _print_with_delay(self, text: str, delay: float):
        try:
            for char in text:
                print(char, end="", flush=True)
                time.sleep(delay)
        except KeyboardInterrupt:
            print()
        finally:
            print()


class PrinterFactory:
    @staticmethod
    def get_printer(style: str) -> Printer:
        if style == "plain":
            return PlainPrinter()
        if style == "markdown":
            return MarkdownPrinter()

        raise ValueError(f"Invalid style: {style}")


class PrintUtils:
    """A bunch of utility messages"""

    INSTALL_WELCOME_MESSAGE = (
        Style.BRIGHT
        + Fore.LIGHTBLUE_EX
        + "\n*~ Welcome to TerminalGPT-Pro installation wizard ~*\n"
        + Style.RESET_ALL
        + """
Please note that you need to have an OpenAI account to use TerminalGPT-Pro and an API key to use TerminalGPT-Pro.
If you don't have an OpenAI account, please create one at https://beta.openai.com/signup.
If you don't have an API key, please create one at https://beta.openai.com/account/api-keys.
"""
        + Style.BRIGHT
        + Fore.LIGHTBLUE_EX
        + """
Let's install the OpenAI API key, so you can use TerminalGPT-Pro.
"""
    )
    # pylint: disable=W1401
    INSTALL_ART = (
        Style.BRIGHT
        + Fore.GREEN
        + """



╔════╗─────────────╔╗╔═══╦═══╦════╗─╔═══╗
║╔╗╔╗║─────────────║║║╔═╗║╔═╗║╔╗╔╗║─║╔═╗║
╚╝║║╠╩═╦═╦╗╔╦╦═╗╔══╣║║║─╚╣╚═╝╠╝║║╚╝─║╚═╝╠═╦══╗
──║║║║═╣╔╣╚╝╠╣╔╗╣╔╗║║║║╔═╣╔══╝─║║╔══╣╔══╣╔╣╔╗║
──║║║║═╣║║║║║║║║║╔╗║╚╣╚╩═║║────║║╚══╣║──║║║╚╝║
──╚╝╚══╩╝╚╩╩╩╩╝╚╩╝╚╩═╩═══╩╝────╚╝───╚╝──╚╝╚══╝
"""
        + Style.RESET_ALL
    )

    INSTALL_SUCCESS_MESSAGE = (
        Style.BRIGHT
        + Fore.GREEN
        + f"""
Great news! You're all set up to use TerminalGPT-Pro!

Your TerminalGPT-Pro files are all saved at {config.BASE_PATH}.
To start chatting with me, just type '{config.APP_NAME}' into your terminal and let the fun begin!

Thanks for choosing TerminalGPT-Pro - the coolest personal assistant on the block.
Let's get our programming on!"""
        + Style.RESET_ALL
    )

    INSTALL_SMALL_PRINTS = (
        """
Just a reminder that TerminalGPT-Pro is a free and open source project.
So if you ever feel like contributing or checking out the inner workings of the program,
feel free to head on over to https://github.com/Plim4ik/TerminalGPT-Pro
Thanks for being a part of our community!

"""
        + Style.RESET_ALL
    )

    CONVERSATIONS_INIT_MESSAGE = (
        Style.BRIGHT
        + Fore.LIGHTBLUE_EX
        + """
Welcome back to TerminalGPT-Pro!
Here are your previous conversations:
"""
        + Style.RESET_ALL
    )

    STOPPED_CONTINUE_MESSAGES = [
        "OK, I stopped. Your wish is my command.",
        "Alright, I stopped. I'm at your service.",
        "OK, I'm all ears. Your wish is my command.",
        "Alright, I paused. I'm at your service.",
        "Alright, I'm waiting for the next instruction.",
        "I've hit pause. Your wish is my command.",
        "Sure thing. Ready and waiting for your next move.",
        "Standing by for your command.",
        "I'm all set, waiting for your orders.",
        "I'm here, awaiting your direction.",
    ]

    STOPPED_MESSAGES = [
        "OK, I stopped. Goodbye.",
        "Alright, I stopped. See you later.",
        "Alright, I paused. Goodbye.",
        "Alright, Goodbye.",
        "I've hit pause. See you later.",
        "Sure thing. See you later.",
    ]

    @staticmethod
    def choose_random_message(messages: List[str]):
        """Chooses a random message from a list of messages."""
        return messages[random.randint(0, len(messages) - 1)]



"""Encryption module for terminalgpt."""

import os
import sys

import openai
from colorama import Fore, Style
from cryptography.fernet import Fernet

from terminalgptpro import config


class EncryptionManager:
    """Manages encryption and decryption of secrets."""

    def __init__(self):
        self.__key_path = config.KEY_PATH

    def set_encryption_key(self):
        """Sets the encryption key for OpenAI API."""

        key = Fernet.generate_key()

        try:
            if not os.path.exists(os.path.dirname(self.__key_path)):
                os.makedirs(os.path.dirname(self.__key_path))

            with open(self.__key_path, "wb") as file:
                file.write(key)
        except OSError:
            print(
                Style.BRIGHT
                + Fore.RED
                + f"""
                Failed to create encryption key at {self.__key_path}.
                Please check your permissions and try again.
                """
                + Style.RESET_ALL
            )
            sys.exit(1)

        return key

    def get_encryption_key(self):
        """Generates a key and save it into a file"""

        key = ""

        try:
            with open(self.__key_path, "rb") as file:
                key = file.read()
        except OSError:
            print(
                Style.BRIGHT
                + Fore.RED
                + f"""
                Failed to read encryption key from {self.__key_path}.
                Please check your permissions and try again.
                """
                + Style.RESET_ALL
            )
            sys.exit(1)

        return key

    def encrypt(self, secret: bytes, key):
        """Encrypts a secret using Fernet encryption."""

        # Create a Fernet cipher using the key
        cipher = Fernet(key)

        # Encrypt the secret
        return cipher.encrypt(secret)

    def decrypt(self, file_path, key):
        """Decrypts a secret using Fernet encryption."""

        # Read the encrypted secret from the file
        with open(file_path, "rb") as file:
            encrypted_secret = file.read()

        # Create a Fernet cipher using the key
        cipher = Fernet(key)

        # Decrypt the secret
        return cipher.decrypt(encrypted_secret).decode()

    def check_api_key(self):
        """Checks if the API key is installed."""

        message = f"""
        OpenAI API key is missing!
        Please install OpenAI API key first with '{config.APP_NAME} install' command.
        """

        if (
            not os.path.exists(config.SECRET_PATH)
            and "OPENAI_API_KEY" not in os.environ
        ):
            print(Style.BRIGHT + Fore.RED + message + Style.RESET_ALL)
            sys.exit(1)

    def set_api_key(self):
        """Sets the API key for OpenAI API."""
        if os.environ.get("OPENAI_API_KEY"):
            openai.api_key = os.environ.get("OPENAI_API_KEY")
        else:
            self.check_api_key()
            key = self.get_encryption_key()
            openai.api_key = self.decrypt(config.SECRET_PATH, key)



import click
from terminalgptpro import config
from terminalgptpro.chat import ChatManager
from terminalgptpro.conversations import ConversationManager
from terminalgptpro.encryption import EncryptionManager
from terminalgptpro.printer import PrinterFactory, PrintUtils

from .terminal.new import start_new_conversation
from .terminal.install import install_api_key
from .terminal.load import load_conversation
from .terminal.delete import delete_conversation
from .terminal.one_shot import one_shot


@click.group()
@click.version_option(prog_name="TerminalGPT-Pro", message="%(prog)s %(version)s")
@click.option(
    "--model",
    "-m",
    type=click.Choice(list(config.MODELS.keys())),
    default=config.get_default_config().get("model", "gpt-3.5-turbo"),
    show_default=True,
    help="Choose a model to use.",
)
@click.option(
    "--style",
    "-s",
    type=click.Choice(["markdown", "plain"]),
    default=config.get_default_config().get("style", "markdown"),
    show_default=True,
    help="Output style.",
)
@click.pass_context
def cli(ctx, model, style: str):
    token_limit = config.MODELS[model]
    safety_buffer = token_limit * 0.25

    ctx.ensure_object(dict)

    ctx.obj["STYLE"] = style
    ctx.obj["PRINTER"] = PrinterFactory.get_printer(style)
    ctx.obj["MODEL"] = model
    ctx.obj["ENC_MNGR"] = EncryptionManager()
    ctx.obj["CONV_MANAGER"] = ConversationManager(ctx.obj["PRINTER"])

    ctx.obj["SESSION"] = PromptSession(
        style=PromptStyle.from_dict({"prompt": "bold"}),
        message="\nUser: ",
    )

    ctx.obj["CHAT"] = ChatManager(
        conversations_manager=ctx.obj["CONV_MANAGER"],
        token_limit=int(token_limit - safety_buffer),
        session=ctx.obj["SESSION"],
        messages=[],
        model=ctx.obj["MODEL"],
        printer=ctx.obj["PRINTER"],
    )

@cli.command(help="Creating a secret api key for the chatbot."
                + " You will be asked to enter your OpenAI API key.")
def install():
    install_api_key()

@cli.command(help="Start a new conversation.")
@click.pass_context
def new(ctx):
    start_new_conversation(ctx)

@cli.command(help="Choose a previous conversation to load.")
@click.pass_context
def load(ctx):
    load_conversation(ctx)

@cli.command(help="Choose a previous conversation to delete.")
@click.pass_context
def delete(ctx):
    delete_conversation(ctx)

# Добавление команды one_shot
@cli.command(help="One shot question answer.")
@click.argument("question", type=str)
@click.pass_context
def one_shot(ctx, question):
    one_shot_message(ctx, question)

cli.add_command(install)
cli.add_command(load)
cli.add_command(new)
cli.add_command(delete)
cli.add_command(one_shot)


if __name__ == "__main__":
    cli()



"""Conversations module."""

import json
import os
import time

import openai
from colorama import Back, Style

from terminalgptpro import config
from terminalgptpro.printer import Printer


class ConversationManager:
    """Manages conversations."""

    def __init__(self, printer: Printer, conversation_name: str = ""):
        self.__base_path = config.CONVERSATIONS_PATH
        self.__conversation_name = conversation_name
        self.__printer = printer

    @property
    def conversation_name(self):
        return self.__conversation_name

    @conversation_name.setter
    def conversation_name(self, conversation_name: str):
        self.__conversation_name = conversation_name

    def get_system_answer(self, messages):
        """Returns the answer from OpenAI API."""

        while True:
            try:
                answer = openai.ChatCompletion.create(
                    model=config.get_default_config()["model"], messages=messages
                )
                return answer
            except openai.OpenAIError:
                time.sleep(10)

    def create_conversation_name(self, messages: list):
        """Creates a context file name based on the title of the conversation."""

        files = self.get_conversations()

        message_suffix = f"- Keep it unique amongst the next file names list: {files}"
        title_message = {
            "role": "system",
            "content": config.TITLE_MESSAGE + message_suffix,
        }

        answer = self.get_system_answer(messages + [title_message])
        context_file_name = answer["choices"][0]["message"]["content"]

        self.conversation_name = context_file_name

    def save_context(self, messages: list, total_usage: int, token_limit: int):
        # Save context or wait for some context
        if not self.conversation_name and total_usage > token_limit / 12:
            self.create_conversation_name(messages)
        elif self.conversation_name:
            self.save_conversation(messages)

    def save_conversation(self, messages: list):
        """Saves a conversation to a file."""

        if not os.path.exists(self.__base_path):
            os.makedirs(self.__base_path)

        with open(
            file=f"{self.__base_path}/{self.conversation_name}",
            mode="w",
            encoding="utf-8",
        ) as conv_file:
            json.dump(messages, conv_file)

    def delete_conversation(self, conversation_name: str):
        """Deletes a conversation from a file."""

        os.remove(f"{self.__base_path}/{conversation_name}")

    def load_conversation(self) -> list:
        """Loads a conversation from a file. returns a list of messages."""

        messages = []

        try:
            with open(
                file=f"{self.__base_path}/{self.conversation_name}",
                mode="r",
                encoding="utf-8",
            ) as conv_file:
                messages = json.load(conv_file)
        except FileNotFoundError as error:
            error_message = f"Failed loading conversation {self.conversation_name} from {self.__base_path}.\n{str(error)}"
            self.__printer.printt(
                Back.RED + Style.BRIGHT + error_message + Style.RESET_ALL
            )
            messages = []

        return messages

    def get_conversations(self):
        """Lists all saved conversations."""

        if not os.path.exists(self.__base_path):
            os.makedirs(self.__base_path)

        files = os.listdir(self.__base_path)
        files.sort(
            key=lambda x: os.path.getmtime(os.path.join(self.__base_path, x)),
            reverse=True,
        )

        return files

    def is_conversations_empty(self, files: list, message: str):
        """Checks if the conversations directory is empty."""

        if files == []:
            self.__printer.printt(message + Style.RESET_ALL)
            return True
        return False



import os
import sys
import time

import openai
import tiktoken
from colorama import Back, Fore, Style
from prompt_toolkit import PromptSession
from tiktoken.core import Encoding
from yaspin import yaspin
from yaspin.spinners import Spinners

from terminalgptpro import config
from terminalgptpro.conversations import ConversationManager
from terminalgptpro.printer import Printer, PrintUtils


class ChatManager:
    def __init__(self, **kwargs):
        self.__tiktoken_encoder: Encoding = tiktoken.get_encoding(config.ENCODING_MODEL)
        self.__convers_manager: ConversationManager = kwargs["conversations_manager"]
        self.__session: PromptSession = kwargs["session"]
        self.__messages: list = kwargs["messages"]
        self.__model = kwargs["model"]
        self.__printer: Printer = kwargs["printer"]
        self.__token_limit: int = kwargs["token_limit"]
        self.__total_usage = 0

    @property
    def messages(self) -> list:
        return self.__messages

    @messages.setter
    def messages(self, messages: list):
        self.__messages = messages

    @property
    def token_limit(self) -> int:
        return self.__token_limit

    @token_limit.setter
    def token_limit(self, token_limit: int):
        self.__token_limit = token_limit

    @property
    def total_usage(self) -> int:
        return self.__total_usage

    @total_usage.setter
    def total_usage(self, total_usage: int):
        self.__total_usage = total_usage

    def __print_usage(self):
        """Prints the total usage"""
        print(
            Fore.LIGHTBLUE_EX
            + f"\nAPI Total Usage: {str(self.__total_usage)} tokens"
            + Style.RESET_ALL
        )
        print(
            Fore.LIGHTCYAN_EX
            + f"Counter Total Usage: {str(self.num_tokens_from_messages())} tokens"
            + Style.RESET_ALL
        )

    def chat_loop(self):
        """Main chat loop."""

        while True:
            # flush stdin
            sys.stdin.flush()

            # Get user input
            user_input = self.__session.prompt()
            self.__printer.printt()

            # Append to messages and send to ChatGPT
            self.__messages.append({"role": "user", "content": user_input})
            self.__total_usage = self.num_tokens_from_messages()

            # Prevent reaching tokens limit
            if self.exceeding_token_limit():
                self.reduce_tokens()

            # Get answer
            try:
                answer = self.get_user_answer(self.__messages)
            except KeyboardInterrupt:
                self.__printer.print_assistant_message(
                    PrintUtils.choose_random_message(
                        PrintUtils.STOPPED_CONTINUE_MESSAGES
                    )
                )
                continue
            except Exception as error:
                self.__printer.print_assistant_message(
                    str(error), color=Back.RED + Style.BRIGHT
                )
                continue

            # Parse total_usage and message from answer
            self.__total_usage = answer["usage"]["total_tokens"]
            message = answer["choices"][0]["message"]["content"]

            # Append to messages list for next iteration keeping context
            self.__messages.append({"role": "assistant", "content": message})

            # Save context or wait for some context
            self.__convers_manager.save_context(
                self.__messages, self.__total_usage, self.__token_limit
            )

            # Print answer message
            self.__printer.print_assistant_message(message)

            # Print usage
            if os.environ.get("LOG_LEVEL") == "DEBUG":
                self.__print_usage()

            if user_input == "exit":
                sys.exit()

    def get_user_answer(self, messages: list):
        """Returns the answer from OpenAI API."""
        while True:
            try:
                with yaspin(
                    Spinners.earth,
                    text=Style.BRIGHT + "Assistant:" + Style.RESET_ALL,
                    color="blue",
                    side="right",
                ):
                    return openai.ChatCompletion.create(
                        model=self.__model, messages=messages
                    )
            except openai.InvalidRequestError as error:
                if "Please reduce the length of the messages" in str(error):
                    self.__messages.pop(1)
                    time.sleep(0.5)
                else:
                    raise error

    def exceeding_token_limit(self):
        """Returns True if the total_usage is greater than the token limit with some safe buffer."""
        return self.__total_usage > self.__token_limit

    def reduce_tokens(self) -> int:
        """Reduce tokens in messages context."""
        total_reduced_amount = self.__total_usage - self.__token_limit
        reduce_amount = self.__total_usage - self.__token_limit
        tokenized_message = []

        while self.exceeding_token_limit():
            message = self.__messages.pop(1)
            tokenized_message = self.__tiktoken_encoder.encode(message["content"])
            while reduce_amount > 0 and len(tokenized_message) > 0:
                self.__total_usage -= 1
                reduce_amount -= 1
                tokenized_message.pop(0)
            if len(tokenized_message) == 0 and self.exceeding_token_limit():
                reduce_amount -= 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
                self.__total_usage -= 4
                for key, _ in message.items():
                    if key == "name":  # if there's a name, the role is omitted
                        reduce_amount += 1  # role is always required and always 1 token
                        self.__total_usage += 1
        if len(tokenized_message) > 0:
            message["content"] = self.__tiktoken_encoder.decode(tokenized_message)
            self.__messages.insert(1, message)
        if os.environ.get("LOG_LEVEL") == "DEBUG":
            self.__print_usage()

        return total_reduced_amount

    def num_tokens_from_messages(self) -> int:
        """Returns the number of tokens used by a list of messages."""
        encoding = tiktoken.get_encoding(config.ENCODING_MODEL)
        num_tokens = 0
        for message in self.__messages:
            num_tokens += (
                4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
            )
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == "name":  # if there's a name, the role is omitted
                    num_tokens -= 1  # role is always required and always 1 token
        num_tokens -= 2  # every reply is primed with <im_start>assistant
        return num_tokens

    def welcome_message(self, messages: list):
        """Prints the welcome message."""
        print()
        try:
            welcome_message = self.get_user_answer(messages)
            self.__printer.print_assistant_message(
                welcome_message["choices"][0]["message"]["content"]
            )
        except KeyboardInterrupt:
            self.__printer.print_assistant_message(
                PrintUtils.choose_random_message(PrintUtils.STOPPED_MESSAGES),
                color=Fore.YELLOW + Style.RESET_ALL,
            )
            sys.exit(0)
        except Exception as error:
            self.__printer.print_assistant_message(
                str(error), color=Back.RED + Style.BRIGHT
            )
            sys.exit(1)



from .config import *
from .chat import *
from .conversations import *
from .encryption import *
from .printer import *
from .main import *


def delete_conversation(ctx):
    """Delete previous conversations."""

    printer: Printer = PrinterFactory.get_printer("plain")
    conv_manager: ConversationManager = ctx.obj["CONV_MANAGER"]
    printer.printt(PrintUtils.CONVERSATIONS_INIT_MESSAGE)

    while True:
        conversations = conv_manager.get_conversations()

        msg = (
            Style.BRIGHT
            + Fore.RED
            + "\n** There are no conversations to delete! **"
            + Style.RESET_ALL
        )

        if conv_manager.is_conversations_empty(files=conversations, message=msg):
            return

        # setup file names auto completion
        completer = WordCompleter(conversations, ignore_case=True)

        # print conversations list
        printer.printt(Style.BRIGHT + "Conversations list:" + Style.RESET_ALL)
        for conversation in conversations:
            printer.printt("- " + conversation)

        conversation = prompt(
            "\nChoose a conversation to delete:\n",
            completer=completer,
            style=PromptStyle.from_dict({"prompt": "bold"}),
        )

        # delete conversation file
        if conversation in conversations:
            conv_manager.delete_conversation(conversation)

            printer.printt(
                Style.BRIGHT
                + Fore.LIGHTBLUE_EX
                + "\n** Conversation "
                + Fore.WHITE
                + conversation
                + Fore.LIGHTBLUE_EX
                + " deleted! **\n"
                + Style.RESET_ALL
            )

            # delete conversation from conversations list
            conversations.remove(conversation)
            completer = WordCompleter(conversations, ignore_case=True)
        else:
            printer.printt(
                Style.BRIGHT
                + Fore.RED
                + "\n** Conversation not found! **"
                + Style.RESET_ALL
            )

        msg = (
            Style.BRIGHT
            + Fore.LIGHTBLUE_EX
            + "\n** No more conversations to delete! **"
            + Style.RESET_ALL
        )
        if conv_manager.is_conversations_empty(files=conversations, message=msg):
            return



from .config import *
from .chat import *
from .conversations import *
from .encryption import *
from .printer import *
from .main import *

def load_conversation(ctx):
    """Load a previous conversation."""

    printer: Printer = PrinterFactory.get_printer("plain")
    chat_manager: ChatManager = ctx.obj["CHAT"]
    enc_manager: EncryptionManager = ctx.obj["ENC_MNGR"]
    conv_manager: ConversationManager = ctx.obj["CONV_MANAGER"]

    messages = []
    enc_manager.set_api_key()

    # get conversations list
    conversations = conv_manager.get_conversations()

    msg = (
        Style.BRIGHT
        + Fore.RED
        + "\n** There are no conversations to load! **"
        + Style.RESET_ALL
    )

    if conv_manager.is_conversations_empty(files=conversations, message=msg):
        return

    # setup file names auto-completion
    completer = WordCompleter(conversations, ignore_case=True)
    printer.printt(PrintUtils.CONVERSATIONS_INIT_MESSAGE)

    # print conversations list
    for conversation in conversations:
        printer.printt(Style.BRIGHT + "- " + conversation)

    # prompt user to choose a conversation and load it into messages
    conversation = prompt(
        "\nChoose a conversation:\n",
        completer=completer,
        style=PromptStyle.from_dict({"prompt": "bold lightblue"}),
    )

    # if conversation not found, return
    if conversation not in conversations:
        printer.printt(
            Style.BRIGHT
            + Fore.RED
            + "\n** Conversation not found! **"
            + Style.RESET_ALL
        )
        return

    # load conversation
    conv_manager.conversation_name = conversation
    while not messages:
        messages = conv_manager.load_conversation()

    messages.append(config.INIT_WELCOME_BACK_MESSAGE)
    chat_manager.messages = messages
    chat_manager.total_usage = chat_manager.num_tokens_from_messages()

    printer.printt(
        Style.BRIGHT
        + "\nConversation details and settings:"
        + Style.RESET_ALL
        + f"""
    Name: {conversation}
    Model: {ctx.obj["MODEL"]}
    Token Limit: {config.MODELS[ctx.obj["MODEL"]]}
    Current Token length: {chat_manager.total_usage}
    """
    )

    if chat_manager.exceeding_token_limit():
        printer.printt(
            Style.BRIGHT
            + Fore.RED
            + "Warning:\n"
            + Style.RESET_ALL
            + "The token length of this conversation is exceeding the token limit (+ some buffer) for the current model.\n"
            "We are about to reduce the token length by removing the oldest messages from the conversation.\n"
        )
        user_approval = prompt(
            "Should we continue? (y/n)  ",
            style=PromptStyle.from_dict({"prompt": "bold lightblue"}),
        )

        if user_approval.lower() != "y":
            printer.printt(
                Style.BRIGHT
                + Fore.LIGHTBLUE_EX
                + "\n** Conversation loading aborted! **\n"
                + Style.RESET_ALL
            )
            return

        printer.printt(
            "\nReducing token length by removing the oldest messages from the conversation...\n"
        )
        chat_manager.reduce_tokens()
        printer.printt(
            f"{Style.BRIGHT}{Fore.GREEN}Token length reduced successfully!{Style.RESET_ALL}"
        )

    printer.printt(
        Style.BRIGHT
        + Fore.LIGHTBLUE_EX
        + "\n** Conversation "
        + Fore.WHITE
        + conversation
        + Fore.LIGHTBLUE_EX
        + " is loaded! **\n"
        + "- - - - - - - - - - - - - - - - - - - - - - - - -"
        + Style.RESET_ALL
    )
    token_limit = int(config.MODELS[ctx.obj["MODEL"]] / 1000)
    printer.printt(
        f"{Style.RESET_ALL}{Style.BRIGHT}Model: {Style.RESET_ALL}{ctx.obj['MODEL']} "
        f"{Style.BRIGHT}Token Limit: {Style.RESET_ALL}{token_limit}k "
        f"{Style.RESET_ALL}{Style.BRIGHT}Style: {Style.RESET_ALL}{ctx.obj['STYLE']}"
    )

    chat_manager.welcome_message(messages=messages)
    messages.pop()

    chat_manager.messages = messages
    chat_manager.chat_loop()